package mvd.jester.model;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import com.google.common.collect.Iterables;
import com.google.common.collect.Sets;
import org.jgrapht.Graphs;
import org.jgrapht.alg.shortestpath.AllDirectedPaths;
import org.jgrapht.experimental.dag.DirectedAcyclicGraph;
import org.jgrapht.graph.DefaultEdge;
import mvd.jester.utils.BinaryDecompositionTree;
import mvd.jester.utils.BinaryDecompositionTree.Node;
import mvd.jester.utils.BinaryDecompositionTree.NodeType;

public class DagTask implements Task {

    private DirectedAcyclicGraph<Job, DefaultEdge> jobDag;
    private final Set<Segment> workloadDistribution;
    private final long workload;
    private final long criticalPath;
    private final long period;
    private final long deadline;
    private final long numberOfThreads;

    public DagTask(DirectedAcyclicGraph<Job, DefaultEdge> jobDag, long period,
            long numberOfThreads) {
        this.jobDag = jobDag;
        this.period = period;
        this.deadline = period;
        this.numberOfThreads = numberOfThreads;
        this.workload = DagUtils.calculateWorkload(this.jobDag);
        this.criticalPath = DagUtils.calculateCriticalPath(this.jobDag);
        this.workloadDistribution =
                DagUtils.calculateWorkloadDistribution(this.jobDag, this.criticalPath);
    }

    public double getUtilization() {
        return (double) workload / period;
    }

    /**
     * @return the deadline
     */
    public long getDeadline() {
        return deadline;
    }

    /**
     * @return the jobDag
     */
    public DirectedAcyclicGraph<Job, DefaultEdge> getJobDag() {
        return jobDag;
    }

    /**
     * @param jobDag the jobDag to set
     */
    public void setJobDag(DirectedAcyclicGraph<Job, DefaultEdge> jobDag) {
        this.jobDag = jobDag;
    }

    /**
     * @return the period
     */
    public long getPeriod() {
        return period;
    }

    /**
     * @return the workload
     */
    public long getWorkload() {
        return workload;
    }

    /**
     * @return the criticalPath
     */
    public long getCriticalPath() {
        return criticalPath;
    }


    /**
     * @return the workloadDistribution
     */
    public Set<Segment> getWorkloadDistribution() {
        return workloadDistribution;
    }

    @Override
    public long getMaximumParallelism() {
        long max = 0;
        for (Segment s : workloadDistribution) {
            if (max < s.getNumberOfJobs()) {
                max = s.getNumberOfJobs();
            }
        }
        return max;
    }


    @Override
    public long getNumberOfThreads() {
        return numberOfThreads;
    }

    public static class DagUtils {
        public static long calculateWorkload(DirectedAcyclicGraph<Job, DefaultEdge> jobDag) {
            long workload = 0;

            for (Job job : jobDag) {
                workload += job.getWcet();
            }
            return workload;
        }

        public static long calculateCriticalPath(DirectedAcyclicGraph<Job, DefaultEdge> jobDag) {
            long criticalPath = 0;
            // BreadthFirstIterator<Job, DefaultEdge> breadthFirstIterator =
            // new BreadthFirstIterator<>(jobDag);
            // while (breadthFirstIterator.hasNext()) {
            // Job job = breadthFirstIterator.next();
            for (Job job : jobDag) {
                Set<DefaultEdge> edges = jobDag.incomingEdgesOf(job);
                long longestRelativeCompletionTime = 0;
                for (DefaultEdge e : edges) {
                    Job source = jobDag.getEdgeSource(e);
                    longestRelativeCompletionTime =
                            longestRelativeCompletionTime >= source.getRelativeCompletionTime()
                                    ? longestRelativeCompletionTime
                                    : source.getRelativeCompletionTime();
                }

                job.setRelativeCompletionTime(longestRelativeCompletionTime + job.getWcet());
                criticalPath = job.getRelativeCompletionTime();
            }

            return criticalPath;
        }

        public static LinkedHashSet<Segment> calculateWorkloadDistribution(
                DirectedAcyclicGraph<Job, DefaultEdge> jobDag, long criticalPath) {
            LinkedHashSet<Segment> segments = new LinkedHashSet<>();
            long segmentDuration = 0;
            long segmentHeight = 1;
            for (long t = 0; t < criticalPath; ++t) {
                long currentHeight = 0;
                for (Job j : jobDag) {
                    if (t >= j.getRelativeCompletionTime() - j.getWcet()
                            && t < j.getRelativeCompletionTime()) {
                        currentHeight++;
                    }
                }
                if (currentHeight == segmentHeight) {
                    segmentDuration++;
                } else {
                    segments.add(new Segment(segmentDuration, segmentHeight));
                    segmentDuration = 1;
                    segmentHeight = currentHeight;
                }
            }
            segments.add(new Segment(segmentDuration, segmentHeight));

            return segments;
        }

        public static DirectedAcyclicGraph<Job, DefaultEdge> createNFJGraph(
                DirectedAcyclicGraph<Job, DefaultEdge> jobDag) {
            LinkedList<Job> joinNodes = new LinkedList<>();
            LinkedList<Job> forkNodes = new LinkedList<>();

            collectForksAndJoins(jobDag, forkNodes, joinNodes);
            return createNFJGraph(jobDag, forkNodes, joinNodes);
        }

        public static void collectForksAndJoins(DirectedAcyclicGraph<Job, DefaultEdge> jobDag,
                LinkedList<Job> forkNodes, LinkedList<Job> joinNodes) {
            for (Job j : jobDag) {
                if (jobDag.inDegreeOf(j) > 1) {
                    joinNodes.add(j);
                }
                if (jobDag.outDegreeOf(j) > 1) {
                    forkNodes.add(j);
                }
            }
        }

        public static DirectedAcyclicGraph<Job, DefaultEdge> createNFJGraph(
                DirectedAcyclicGraph<Job, DefaultEdge> jobDag, LinkedList<Job> forkNodes,
                LinkedList<Job> joinNodes) {
            DirectedAcyclicGraph<Job, DefaultEdge> modifiedJobDag =
                    new DirectedAcyclicGraph<>(DefaultEdge.class);
            Graphs.addGraph(modifiedJobDag, jobDag);

            if (!(joinNodes.size() > 1)) {
                return modifiedJobDag;
            }

            final Job sink = joinNodes.getLast();

            for (Job j : joinNodes) {
                Set<DefaultEdge> edgeSet = new HashSet<>(modifiedJobDag.incomingEdgesOf(j));
                for (DefaultEdge e : edgeSet) {
                    Job predecessor = modifiedJobDag.getEdgeSource(e);
                    boolean satisfiesProposition =
                            DagUtils.checkForFork(modifiedJobDag, j, forkNodes, predecessor);
                    if (!satisfiesProposition) {
                        modifiedJobDag.removeEdge(e);
                        if (modifiedJobDag.outgoingEdgesOf(predecessor).isEmpty()) {
                            try {
                                modifiedJobDag.addDagEdge(predecessor, sink);
                            } catch (Exception ex) {
                            }
                        }
                    }

                    if (modifiedJobDag.inDegreeOf(j) == 1) {
                        break;
                    }
                }
            }

            return modifiedJobDag;
        }

        public static BinaryDecompositionTree<Job> createDecompositionTree(
                DirectedAcyclicGraph<Job, DefaultEdge> jobDag) {
            LinkedList<Job> joinNodes = new LinkedList<>();
            LinkedList<Job> forkNodes = new LinkedList<>();

            collectForksAndJoins(jobDag, forkNodes, joinNodes);
            return createDecompositionTree(jobDag, forkNodes, joinNodes);
        }

        public static BinaryDecompositionTree<Job> createDecompositionTree(
                DirectedAcyclicGraph<Job, DefaultEdge> jobDag, LinkedList<Job> forkNodes,
                LinkedList<Job> joinNodes) {
            Job source = forkNodes.getFirst();
            Job sink = joinNodes.getLast();
            Job current = source;

            BinaryDecompositionTree<Job> tree = new BinaryDecompositionTree<>();
            DagUtils.constructTree(jobDag, null, current, tree, forkNodes, joinNodes, sink);

            return tree;
        }

        private static Node<Job> constructTree(DirectedAcyclicGraph<Job, DefaultEdge> jobDag,
                Node<Job> parent, Job current, BinaryDecompositionTree<Job> tree,
                LinkedList<Job> forkVertices, LinkedList<Job> joinVertices, Job sink) {
            if (forkVertices.contains(current)) {
                Job join = findJoin(jobDag, current, sink, joinVertices);
                if (jobDag.outDegreeOf(join) > 1) {
                    throw new RuntimeException("Join und Fork in Einem! Ohoh!");
                }

                Node<Job> forkNode = new Node<Job>(parent, NodeType.SEQ);
                if (tree.isEmpty()) {
                    tree.setRoot(forkNode);
                }
                forkNode.setLeftNode(new Node<Job>(forkNode, current));
                final Node<Job> continuationNode;
                if (!tree.contains(join)) {
                    forkNode.setRightNode(new Node<Job>(forkNode, NodeType.SEQ));
                    Node<Job> joinNode = forkNode.getRightNode();
                    Iterator<Job> successorIterator =
                            Graphs.successorListOf(jobDag, join).iterator();
                    if (successorIterator.hasNext()) {
                        Job successor = successorIterator.next();
                        if (!tree.contains(successor)) {
                            joinNode.setRightNode(new Node<>(joinNode, NodeType.SEQ));
                            Node<Job> successorNode = joinNode.getRightNode();
                            successorNode.setLeftNode(new Node<>(successorNode, join));
                            Node<Job> subTree = constructTree(jobDag, successorNode, successor,
                                    tree, forkVertices, joinVertices, sink);
                            successorNode.setRightNode(subTree);
                        } else {
                            joinNode.setRightNode(new Node<>(joinNode, join));
                        }
                    } else {
                        joinNode.setRightNode(new Node<>(joinNode, join));
                    }
                    joinNode.setLeftNode(new Node<>(joinNode, NodeType.PAR));
                    continuationNode = joinNode.getLeftNode();
                } else {
                    forkNode.setRightNode(new Node<>(forkNode, NodeType.PAR));
                    continuationNode = forkNode.getRightNode();
                }

                Node<Job> successorParent = continuationNode;
                List<Job> successors = Graphs.successorListOf(jobDag, current);
                // create leftSide of joinNode
                for (int i = 0; i < successors.size(); ++i) {
                    Job succ = successors.get(i);
                    Node<Job> thisNode = constructTree(jobDag, successorParent, succ, tree,
                            forkVertices, joinVertices, sink);
                    if (i == successors.size() - 1) {
                        successorParent.setRightNode(thisNode);
                    } else if (i == successors.size() - 2) {
                        successorParent.setLeftNode(thisNode);
                    } else {
                        successorParent.setLeftNode(thisNode);
                        successorParent.setRightNode(new Node<>(successorParent, NodeType.PAR));
                        successorParent = successorParent.getRightNode();
                    }
                }

                return forkNode;
            } else {
                return new Node<>(parent, current);
            }
        }

        public static boolean checkNFJProperty(DirectedAcyclicGraph<Job, DefaultEdge> jobDag) {
            LinkedList<Job> joinNodes = new LinkedList<>();
            LinkedList<Job> forkNodes = new LinkedList<>();

            collectForksAndJoins(jobDag, forkNodes, joinNodes);

            if (!(joinNodes.size() > 1)) {
                return true;
            }

            for (Job j : joinNodes) {
                nextFork: for (Job f : forkNodes) {
                    Set<DefaultEdge> edgeSet = jobDag.getAllEdges(f, j);

                    for (DefaultEdge e : edgeSet) {
                        Job a = jobDag.getEdgeSource(e);
                        if (a != f) {
                            Set<Job> succAndPred = new HashSet<>();
                            succAndPred.addAll(Graphs.predecessorListOf(jobDag, a));
                            succAndPred.addAll(Graphs.successorListOf(jobDag, a));

                            for (Job b : succAndPred) {
                                if (!((jobDag.getAncestors(jobDag, j).contains(b) || b == j)
                                        && (jobDag.getDescendants(jobDag, f).contains(b)
                                                || b == f))) {
                                    continue nextFork;
                                }
                            }
                        }
                    }
                    return true;
                }
            }
            return false;
        }

        private static Job findJoin(DirectedAcyclicGraph<Job, DefaultEdge> jobDag, Job forkNode,
                Job sink, List<Job> joinNodes) {
            for (Job j : joinNodes) {
                final Set<DefaultEdge> edgeSet = new HashSet<>();
                AllDirectedPaths<Job, DefaultEdge> finder = new AllDirectedPaths<>(jobDag);
                finder.getAllPaths(forkNode, j, true, null)
                        .forEach(g -> edgeSet.addAll(g.getEdgeList()));
                Set<DefaultEdge> outgoingEdges = jobDag.outgoingEdgesOf(forkNode);
                if (edgeSet.containsAll(outgoingEdges)) {
                    return j;
                }
            }

            return sink;
        }

        private static boolean checkForFork(DirectedAcyclicGraph<Job, DefaultEdge> jobDag,
                Job joinNode, List<Job> forkNodes, Job job) {
            List<Job> pred = Graphs.predecessorListOf(jobDag, job);

            for (Job p : pred) {
                if (forkNodes.contains(p)) {
                    for (Job successor : Graphs.successorListOf(jobDag, p)) {
                        if (jobDag.getAncestors(jobDag, joinNode).contains(successor)) {
                            return false;
                        }
                    }
                } else {
                    return checkForFork(jobDag, joinNode, forkNodes, p);
                }
            }

            return true;
        }


        public static DirectedAcyclicGraph<Job, DefaultEdge> createNFJfromDecompositionTree(
                BinaryDecompositionTree<Job> tree) {
            DirectedAcyclicGraph<Job, DefaultEdge> jobDag =
                    new DirectedAcyclicGraph<>(DefaultEdge.class);

            traverseNodes(jobDag, tree.getRootNode(), TraversalOrder.LEFT);

            return jobDag;

        }

        private static Set<Job> traverseNodes(DirectedAcyclicGraph<Job, DefaultEdge> jobDag,
                Node<Job> node, TraversalOrder order) {
            switch (node.getNodeType()) {
                case LEAF: {
                    Job j = node.getObject();
                    jobDag.addVertex(j);
                    return new HashSet<Job>(Arrays.asList(j));
                }
                case SEQ: {
                    Set<Job> left = traverseNodes(jobDag, node.getLeftNode(), TraversalOrder.LEFT);
                    Set<Job> right =
                            traverseNodes(jobDag, node.getRightNode(), TraversalOrder.RIGHT);
                    try {
                        for (Job l : left) {
                            for (Job r : right) {
                                jobDag.addDagEdge(l, r);
                            }
                        }
                    } catch (Exception e) {
                    }
                    if (order == TraversalOrder.LEFT) {
                        Set<Job> openEndJobs = new HashSet<>();
                        for (Job j : right) {
                            if (jobDag.outDegreeOf(j) == 0) {
                                openEndJobs.add(j);
                            }
                            for (Job d : jobDag.getDescendants(jobDag, j)) {
                                if (jobDag.outDegreeOf(d) == 0) {
                                    openEndJobs.add(d);
                                }
                            }
                        }
                        return openEndJobs;
                    } else {
                        return left;
                    }
                    // return order == TraversalOrder.LEFT ? right : left; // left -> right; right
                    // ->
                }
                case PAR: {
                    Set<Job> left = traverseNodes(jobDag, node.getLeftNode(), order);
                    Set<Job> right = traverseNodes(jobDag, node.getRightNode(), order);
                    return Sets.newHashSet(Iterables.concat(left, right));
                }
                default:
                    break;
            }
            return new HashSet<>();
        }

        private class GraphEndPoints {
            public Set<Job> left;
            public Set<Job> right;

            public GraphEndPoints(Set<Job> left, Set<Job> right) {
                this.left = left;
                this.right = right;
            }
        }

        private enum TraversalOrder {
            LEFT, RIGHT
        }

    }
}
