package mvd.jester.model;

import static org.junit.jupiter.api.Assertions.assertTrue;
import org.jgrapht.experimental.dag.DirectedAcyclicGraph;
import org.jgrapht.graph.DefaultEdge;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import mvd.jester.model.DagTask.DagUtils;
import mvd.jester.model.SystemSetup.DagTaskBuilder;
import mvd.jester.utils.BinaryDecompositionTree;

public class TestDagUtils {

    @Test
    @DisplayName("Test if a NFJ Graph is created correctly.")
    public void checkNFJCreation() {
        for (int i = 0; i < 100; ++i) {
            DagTaskBuilder builder = new DagTaskBuilder();
            DirectedAcyclicGraph<Job, DefaultEdge> jobDag = builder.generateTask().getJobDag();

            DirectedAcyclicGraph<Job, DefaultEdge> modifiedJobDag = DagUtils.createNFJGraph(jobDag);

            assertTrue(DagUtils.checkNFJProperty(modifiedJobDag));
        }
    }


    @Test
    @DisplayName("Test if dynamic segments are constructed correctly.")
    public void checkPathLength() {
        DirectedAcyclicGraph<Job, DefaultEdge> jobDag = createJobDag();
        long criticalPath = DagUtils.calculateCriticalPath(jobDag);

        for (Job j : jobDag) {
            assertTrue(j.getRelativeCompletionTime() <= criticalPath);
        }
        for (DefaultEdge e : jobDag.edgeSet()) {
            Job source = jobDag.getEdgeSource(e);
            Job target = jobDag.getEdgeTarget(e);
            assertTrue(source.getRelativeCompletionTime() < target.getRelativeCompletionTime());
        }

        assertTrue(criticalPath == 155);
    }


    @Test
    @DisplayName("Check if Decomposition Tree is created correctly for NFJ Dag.")
    void checkDecompositionTreeCreation() {
        DirectedAcyclicGraph<Job, DefaultEdge> jobDag = createJobDag();

        BinaryDecompositionTree<Job> tree = DagUtils.createDecompositionTree(jobDag);
        DirectedAcyclicGraph<Job, DefaultEdge> jobDagFromTree =
                DagUtils.createNFJfromDecompositionTree(tree);

        assertTrue(jobDag.equals(jobDagFromTree));
    }


    private DirectedAcyclicGraph<Job, DefaultEdge> createJobDag() {
        DirectedAcyclicGraph<Job, DefaultEdge> jobDag =
                new DirectedAcyclicGraph<>(DefaultEdge.class);

        Job source = new Job(20);
        Job pathA = new Job(10);
        Job pathB = new Job(30);
        Job pathC = new Job(40);

        Job child1OfA = new Job(5);
        Job child2OfA = new Job(15);
        Job child3OfA = new Job(5);

        Job childsOfAJoin = new Job(10);

        Job child1OfC = new Job(5);
        Job child2OfC = new Job(10);
        Job child3OfC = new Job(15);
        Job child4OfC = new Job(20);

        Job firstJoin = new Job(20);

        Job intermediateFork = new Job(10);

        Job forkChild1 = new Job(5);
        Job forkChild2 = new Job(10);
        Job forkChild3 = new Job(15);

        Job sink = new Job(30);
        try {
            jobDag.addVertex(source);
            jobDag.addVertex(pathA);
            jobDag.addVertex(pathB);
            jobDag.addVertex(pathC);
            jobDag.addVertex(child1OfA);
            jobDag.addVertex(child2OfA);
            jobDag.addVertex(child3OfA);
            jobDag.addVertex(childsOfAJoin);
            jobDag.addVertex(child1OfC);
            jobDag.addVertex(child2OfC);
            jobDag.addVertex(child3OfC);
            jobDag.addVertex(child4OfC);
            jobDag.addVertex(firstJoin);
            jobDag.addVertex(intermediateFork);
            jobDag.addVertex(forkChild1);
            jobDag.addVertex(forkChild2);
            jobDag.addVertex(forkChild3);
            jobDag.addVertex(sink);
            jobDag.addDagEdge(source, pathA);
            jobDag.addDagEdge(pathA, child1OfA);
            jobDag.addDagEdge(pathA, child2OfA);
            jobDag.addDagEdge(pathA, child3OfA);
            jobDag.addDagEdge(child1OfA, childsOfAJoin);
            jobDag.addDagEdge(child2OfA, childsOfAJoin);
            jobDag.addDagEdge(child3OfA, childsOfAJoin);
            jobDag.addDagEdge(childsOfAJoin, firstJoin);
            jobDag.addDagEdge(source, pathB);
            jobDag.addDagEdge(pathB, firstJoin);
            jobDag.addDagEdge(source, pathC);
            jobDag.addDagEdge(pathC, child1OfC);
            jobDag.addDagEdge(pathC, child2OfC);
            jobDag.addDagEdge(pathC, child3OfC);
            jobDag.addDagEdge(pathC, child4OfC);
            jobDag.addDagEdge(child1OfC, firstJoin);
            jobDag.addDagEdge(child2OfC, firstJoin);
            jobDag.addDagEdge(child3OfC, firstJoin);
            jobDag.addDagEdge(child4OfC, firstJoin);
            jobDag.addDagEdge(firstJoin, intermediateFork);
            jobDag.addDagEdge(intermediateFork, forkChild1);
            jobDag.addDagEdge(intermediateFork, forkChild2);
            jobDag.addDagEdge(intermediateFork, forkChild3);
            jobDag.addDagEdge(forkChild1, sink);
            jobDag.addDagEdge(forkChild2, sink);
            jobDag.addDagEdge(forkChild3, sink);
        } catch (Exception e) {
        }
        return jobDag;
    }
}
